@@ -1,4 +1,4 @@
-# This code is part of Perl distribution String-Print version 1.00.
+# This code is part of Perl distribution String-Print version 1.02.
 # The POD got stripped from this file by OODoc version 3.05.
 # For contributors see file ChangeLog.
 
@@ -10,7 +10,7 @@
 
 
 package String::Print;{
-our $VERSION = '1.00';
+our $VERSION = '1.02';
 }
 
 
@@ -21,6 +21,7 @@ use utf8;
 #use Log::Report::Optional 'log-report';
 
 use Unicode::GCString ();
+use Data::Dumper      ();
 use Date::Parse       qw/str2time/;
 use Encode            qw/is_utf8 decode/;
 use HTML::Entities    qw/encode_entities/;
@@ -32,21 +33,38 @@ my @default_modifiers   = (
 	qr/BYTES\b/      => \&_modif_bytes,
 	qr/HTML\b/       => \&_modif_html,
 	qr/YEAR\b/       => \&_modif_year,
-	qr/DT\([^)]*\)/  => \&_modif_dt,
-	qr/DT\b/         => \&_modif_dt,
-	qr/DATE\b/       => \&_modif_date,
 	qr/TIME\b/       => \&_modif_time,
 	qr/\=/           => \&_modif_name,
+	qr/DATE\([^)]*\)|DATE\b/ => \&_modif_date,
+	qr/DT\([^)]*\)|DT\b/     => \&_modif_dt,
+	qr!UNKNOWN\([0-9]+\)|UNKNOWN\b!       => \&_modif_unknown,
 	qr!CHOP\([0-9]+(?:\,?[^)]*)\)|CHOP\b! => \&_modif_chop,
 	qr!EL\([0-9]+(?:\,?[^)]*)\)|EL\b!     => \&_modif_ellipsis,
 	qr!//(?:\"[^"]*\"|\'[^']*\'|\w+)!     => \&_modif_undef,
 );
 
+# Be warned: %F and %T (from C99) are not always supported on Windows
+my %dt_format = (
+	ASC     => '%a %b %e %H:%M:%S %Y',
+	ISO     => '%Y-%m-%dT%H:%M:%S%z',
+	RFC822  => '%a, %d %b %y %H:%M:%S %z',
+	RFC2822 => '%a, %d %b %Y %H:%M:%S %z',
+	RFC5322 => '%a, %d %b %Y %H:%M:%S %z',
+	FT      => '%Y-%m-%d %H:%M:%S',
+);
+
+my %date_format = (
+	'-'     => '%Y-%m-%d',
+	'/'     => '%Y/%m/%d',
+);
+
 my %defaults = (
-	CHOP   => +{ width => 20, head => '[', units => '', tail => ']' },
-	DT     => +{ standard => 'FT' },
-	EL     => +{ width => 20, replace => '⋯ '},
-	FORMAT => +{ thousands => '' },
+	CHOP      => +{ width => 30, head => '[', units => '', tail => ']' },
+	DATE      => +{ format => $date_format{'-'}, },
+	DT        => +{ format => $dt_format{FT}, },
+	EL        => +{ width  => 30, replace => '⋯ '},
+	FORMAT    => +{ thousands => '' },
+	UNKNOWN   => +{ width  => 30, trim => 'EL' },
 );
 
 my %default_serializers = (
@@ -329,7 +347,7 @@ sub _modif_format_d($$$$)
 {	my ($value, $padding, $max, $sep) = @_;
 	my $d = sprintf "%d", $value;   # what perl usually does with floats etc
 	my $v = length $sep ? reverse(reverse($d) =~ s/([0-9][0-9][0-9])/$1$sep/gr) : $d;
-	$v =~ s/^\.//;
+	$v =~ s/^\Q$sep//;
 
 	if($d !~ /^\-/)
 	{	$v = "+$v" if $padding eq '+';
@@ -372,7 +390,7 @@ sub _modif_bytes($$$)
 {	my ($self, $format, $value, $args) = @_;
 	defined $value && length $value or return undef;
 
-	return sprintf("%3d  B", $value) if $value < 1000;
+	return sprintf("%3d B", $value) if $value < 1000;
 
 	my @scale = qw/kB MB GB TB PB EB ZB/;
 	$value /= 1024;
@@ -382,10 +400,10 @@ sub _modif_bytes($$$)
 		$value /= 1024;
 	}
 
-	return sprintf "%3d $scale[0]", $value + 0.5
+	return sprintf "%3d$scale[0]", $value + 0.5
 		if $value > 9.949;
 
-	sprintf "%3.1f $scale[0]", $value;
+	sprintf "%3.1f$scale[0]", $value;
 }
 
 sub _modif_html($$$)
@@ -393,16 +411,6 @@ sub _modif_html($$$)
 	defined $value ? (encode_entities $value) : undef;
 }
 
-# Be warned: %F and %T (from C99) are not supported on Windows
-my %dt_format = (
-	ASC     => '%a %b %e %T %Y',
-	ISO     => '%Y-%m-%dT%T%z',
-	RFC822  => '%a, %d %b %y %T %z',
-	RFC2822 => '%a, %d %b %Y %T %z',
-	RFC5322 => '%a, %d %b %Y %T %z',
-	FT      => '%Y-%m-%d %T',
-);
-
 sub _modif_year($$$)
 {	my ($self, $format, $value, $args) = @_;
 	defined $value or return undef;
@@ -425,19 +433,29 @@ sub _modif_date($$$)
 {	my ($self, $format, $value, $args) = @_;
 	defined $value or return undef;
 
-	blessed $value && $value->isa('DateTime')
-		and return $value->ymd;
+	my $defaults = $self->defaults('DATE');
+	my $kind     = ($format =~ m/^DATE\(([^)]*)\)/ ? $1 : undef) || $defaults->{format};
+	my $pattern  = $date_format{$kind} // $kind;
 
-	length $value or return undef;
+	my ($y, $m, $d);
+	if(blessed $value && $value->isa('DateTime'))
+	{	($y, $m, $d) = ($value->year, $value->month, $value->day);
+	}
+	elsif( $value =~ m!^\s*([0-9]{4})[:/.-]([0-9]?[0-9])[:/.-]([0-9]?[0-9])\s*$!
+		|| $value =~ m!^\s*([0-9]{4})([0-9][0-9])([0-9][0-9])\s*$!)
+	{	($y, $m, $d) = ($1, $2, $3);
+	}
+	else
+	{	my $stamp = $value =~ /\D/ ? str2time($value) : $value;
+		defined $stamp or return "date not found in '$value'";
+		($y, $m, $d) = (localtime $stamp)[5, 4, 3];
+		$y += 1900; $m++;
+	}
 
-	return sprintf("%4d-%02d-%02d", $1, $2, $3)
-		if $value =~ m!^\s*([0-9]{4})[:/.-]([0-9]?[0-9])[:/.-]([0-9]?[0-9])\s*$!
-		|| $value =~ m!^\s*([0-9]{4})([0-9][0-9])([0-9][0-9])\s*$!;
-
-	my $stamp = $value =~ /\D/ ? str2time($value) : $value;
-	defined $stamp or return "date not found in '$value'";
-
-	strftime "%Y-%m-%d", localtime($stamp);
+	$pattern
+		=~ s/\%Y/$y/r
+		=~ s/\%m/sprintf "%02d", $m/re
+		=~ s/\%d/sprintf "%02d", $d/re;
 }
 
 sub _modif_time($$$)
@@ -469,10 +487,8 @@ sub _modif_dt($$$)
 	length $value or return undef;
 
 	my $defaults = $self->defaults('DT');
-	my $kind     = ($format =~ m/^DT\(([^)]*)\)/ ? $1 : undef) || $defaults->{standard};
-
-	my $pattern  = $dt_format{$kind}
-		or return "dt format $kind not known";
+	my $kind     = ($format =~ m/^DT\(([^)]*)\)/ ? $1 : undef) || $defaults->{format};
+	my $pattern  = $dt_format{$kind} // $kind;
 
 	my $stamp = $value =~ /\D/ ? str2time($value) : $value;
 	defined $stamp or return "dt not found in '$value'";
@@ -494,10 +510,11 @@ sub _modif_name($$$)
 sub _modif_chop($$$)
 {	my ($self, $format, $value, $args) = @_;
 	defined $value && length $value or return undef;
-	$format =~ m/^ CHOP\( ([0-9]+) \,? ([^)]+)? \) | CHOP\b /x or die $format;
 
 	my $defaults = $self->defaults('CHOP');
-	my ($width, $units) = ($1 // $defaults->{width}, $2 // $defaults->{units});
+	$format =~ m/^ CHOP\( ([0-9]+) \,? ([^)]+)? \) | CHOP\b /x or die $format;
+	my $width    = $1 // $args->{width} // $defaults->{width};
+	my $units    = $2 // $args->{units} // $defaults->{units};
 	$width != 0 or return $value;
 
 	# max width of a char is 2
@@ -533,10 +550,11 @@ sub _modif_chop($$$)
 sub _modif_ellipsis($$$)
 {	my ($self, $format, $value, $args) = @_;
 	defined $value && length $value or return undef;
-	$format =~ m/^ EL\( ([0-9]+) \,? ([^)]+)? \) | EL\b /x or die $format;
 
 	my $defaults = $self->defaults('EL');
-	my ($width, $replace) = ($1 // $defaults->{width}, $2 // $defaults->{replace});
+	$format =~ m/^ EL\( ([0-9]+) \,? ([^)]+)? \) | EL\b /x or die $format;
+	my $width    = $1 // $args->{width}   // $defaults->{width};
+	my $replace  = $2 // $args->{replace} // $defaults->{replace};
 	$width != 0 or return $value;
 
 	# max width of a char is 2
@@ -557,6 +575,29 @@ sub _modif_ellipsis($$$)
 	$v->as_string . $pad . $replace;
 }
 
+sub _modif_unknown($$$)
+{	my ($self, $format, $value, $args) = @_;
+	defined $value or return undef;
+
+	my $defaults = $self->defaults('UNKNOWN');
+	$format =~ m/^ UNKNOWN\( ([0-9]+) \) | UNKNOWN\b /x or die $format;
+	$args->{width} = $1 // $args->{width} // $defaults->{width};
+
+	return ref $value
+		if blessed $value;
+
+	my $trim    = $args->{trim} // $defaults->{trim};
+	my $trimmer = $trim eq 'EL' ? '_modif_ellipsis' : $trim eq 'CHOP' ? '_modif_chop' : die $trim;
+	my $shorten = sub { $self->$trimmer($trim, $_[0], $args) };
+
+	my $serial  = Data::Dumper->new([$value])->Quotekeys(0)->Terse(1)->Useqq(1)->Indent(0)->Sortkeys(1)->Dump;
+
+	  ! reftype $value          ?  '"' . $shorten->($serial =~ s/^\"//r =~ s/\"$//r) . '"'
+	: reftype $value eq 'ARRAY' ?  '[' . $shorten->($serial =~ s/^\[//r =~ s/\]$//r) . ']'
+	: reftype $value eq 'HASH'  ?  '{' . $shorten->($serial =~ s/^\{//r =~ s/\}$//r) . '}'
+	:     $shorten->($serial);
+}
+
 
 sub printi($$@)
 {	my $self = shift;
